# Copyright [2020] [Toyota Research Institute]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Module and script for validating battery cycler flat files using dictionary
key or DataFrame based validation of typing, min/max, and non-allowed values

Usage:
    validate [INPUT_JSON]

Options:
    -h --help       Show this screen
    --version       Show version

The validation script, `validate`, runs the validation procedure contained
in beep.validate on renamed files according to the output of `collate`.
It also updates a general json validation record in `/data-share/validation/validation.json`.

The input json must contain the following fields

* `file_list` - the list of filenames to be validated

The output json will have the following fields:

* `validity` - a list of boolean validation results, e. g. `[True, True, False]`
* `file_list` - a list of full path filenames which have been processed

Example:
$ validate '{"fid": [0, 1, 2], "strname": ["2017-05-09_test-TC-contact", "2017-08-14_8C-5per_3_47C",
    "2017-12-04_4_65C-69per_6C"], "file_list": ["/data-share/renamed_cycler_files/FastCharge/FastCharge_0_CH33.csv",
    "/data-share/renamed_cycler_files/FastCharge/FastCharge_1_CH44.csv",
    "/data-share/renamed_cycler_files/FastCharge/FastCharge_2_CH29.csv"],
    "protocol": [null, "8C(5%)-3.47C", "4.65C(69%)-6C"], "date": ["2017-05-09", "2017-08-14", "2017-12-04"],
    "channel_no": ["CH33", "CH44", "CH29"],
    "filename": ["/data-share/raw_cycler_files/2017-05-09_test-TC-contact_CH33.csv",
        "/data-share/raw_cycler_files/2017-08-14_8C-5per_3_47C_CH44.csv",
        "/data-share/raw_cycler_files/2017-12-04_4_65C-69per_6C_CH29.csv"]}'
{"validity": [false, false, true], "file_list": ["/data-share/renamed_cycler_files/FastCharge/FastCharge_0_CH33.csv",
    "/data-share/renamed_cycler_files/FastCharge/FastCharge_1_CH44.csv",
    "/data-share/renamed_cycler_files/FastCharge/FastCharge_2_CH29.csv"]}
"""

import json
import os
import warnings
import re
from datetime import datetime

import numpy as np
import pandas as pd
from docopt import docopt
from cerberus import Validator
from beep import tqdm
from monty.serialization import loadfn, dumpfn

from beep import VALIDATION_SCHEMA_DIR
from beep.conversion_schemas import ARBIN_CONFIG, MACCOR_CONFIG
from beep.utils import WorkflowOutputs
from beep import logger, __version__

DEFAULT_ARBIN_SCHEMA = os.path.join(VALIDATION_SCHEMA_DIR, "schema-arbin-lfp.yaml")
CERBERUS_ARBIN_SCHEMA = os.path.join(VALIDATION_SCHEMA_DIR, "schema-arbin-cerberus.yaml")
DEFAULT_MACCOR_SCHEMA = os.path.join(VALIDATION_SCHEMA_DIR, "schema-maccor-lfp.yaml")
DEFAULT_EIS_SCHEMA = os.path.join(VALIDATION_SCHEMA_DIR, "schema-maccor-eis.yaml")
PROJECT_SCHEMA = os.path.join(VALIDATION_SCHEMA_DIR, "schema-projects.yaml")
DEFAULT_VALIDATION_RECORDS = os.path.join(
    VALIDATION_SCHEMA_DIR, "validation_records.json"
)
s = {"service": "DataValidator"}


class ValidatorBeep(Validator):
    """
    Data validation for battery cycling.
    Currently supports Arbin and Maccor cyclers.

    """

    def validate_arbin_dataframe(self, df, schema=CERBERUS_ARBIN_SCHEMA):
        """
        Validator for large, cyclic dataframes coming from Arbin.
        Requires a valid Cycle_Index column of type int.
        Designed for performance - will stop at the first encounter of issues.

        Args:
            df (pandas.DataFrame): Arbin output as DataFrame.
            schema (str): Path to the validation schema. Defaults to arbin for now.
        Returns:
            bool: True if validated with out errors. If validation fails, errors
                are listed at ValidatorBeep.errors.
        """

        try:
            schema = loadfn(schema)
            self.arbin_schema = schema
        except Exception as e:
            warnings.warn("Arbin schema could not be found: {}".format(e))

        df = df.rename(str.lower, axis="columns")

        # Validation cycle index data and cast to int
        if not self._prevalidate_nonnull_column(df, "cycle_index"):
            return False
        df.cycle_index = df.cycle_index.astype(int, copy=False)

        # Validation starts here
        self.schema = self.arbin_schema

        for cycle_index, cycle_df in tqdm(df.groupby("cycle_index")):
            cycle_dict = cycle_df.replace({np.nan, "None"}).to_dict(orient="list")
            result = self.validate(cycle_dict)
            if not result:
                return False
        return True

    def validate_maccor_dataframe(self, df, schema=DEFAULT_MACCOR_SCHEMA):
        """
        Validator for large, cyclic dataframes coming from Maccor.
        Requires a valid Cyc# column of type int.
        Designed for performance - will stop at the first encounter of issues.

        Args:
            df (pandas.DataFrame): Maccor output as DataFrame.
            schema (str): Path to the validation schema. Defaults to maccor for now.
        Returns:
            bool: True if validated with out errors. If validation fails, errors
            are listed at ValidatorBeep.errors.
        """

        try:
            schema = loadfn(schema)
            self.maccor_schema = schema
        except Exception as e:
            warnings.warn("Maccor schema could not be found: {}".format(e))

        df = df.rename(str.lower, axis="columns")

        # Validation cycle index data and cast to int
        if not self._prevalidate_nonnull_column(df, "cyc#"):
            return False
        df["cyc#"] = df["cyc#"].astype(int, copy=False)

        # Validation starts here
        self.schema = self.maccor_schema

        for cycle_index, cycle_df in tqdm(df.groupby("cyc#")):
            cycle_dict = cycle_df.replace({np.nan, "None"}).to_dict(orient="list")
            result = self.validate(cycle_dict)
            if not result:
                return False
        return True

    def validate_eis_dataframe(self, df, schema=DEFAULT_EIS_SCHEMA):
        """
        Validator for Maccor EIS.
        Args:
            df (pandas.DataFrame): Maccor EIS output as DataFrame.
            schema (str): Path to the validation schema. Defaults to maccor for now.
        Returns:
            bool: True if validated with out errors. If validation fails, errors
                are listed at ValidatorBeep.errors.
        """

        try:
            schema = loadfn(schema)
            self.eis_schema = schema
        except Exception as e:
            warnings.warn("Maccor EIS schema could not be found: {}".format(e))

        df = df.rename(str.lower, axis="columns")
        self.schema = self.eis_schema

        return self.validate(df.replace(np.nan, "None").to_dict(orient="list"),)

    def validate_from_paths(
        self,
        paths,
        record_results=False,
        skip_existing=False,
        record_path=DEFAULT_VALIDATION_RECORDS,
    ):
        """
        This method streamlines validation of multiple Arbin csv files given a list of paths.

        It can also do bookkeeping of validations by dumping results in a json file,
        locally until a more centralized method is implemented.

        Args:
            paths (list): a list of paths to csv files.
            record_results (bool): Whether to record the validation results locally or not (defaults to False).
            skip_existing (bool): Whether to skip already validated files. This is done by checking if the file is in
                the validation_records. skip_existing only matters if record_results is True. Defaults to False.
            record_path (str): path to the json file storing the past validation results.
        Returns:
            dict: Results of the validation in the form of a key,value pairs where each key corresponds to the filename
                validated. For each file, the results contain a field "validated", True if validation was successful or
                False if not. "errors", "method" and "time" are simply the errors encountered during validation, method
                used for validation, and time of validation, respectively.

        """
        self.allow_unknown = True

        if record_results:
            if os.path.isfile(record_path):
                self.validation_records = loadfn(record_path)
                if skip_existing:
                    paths = [
                        path
                        for path in paths
                        if os.path.basename(path) not in self.validation_records
                    ]
            else:
                self.validation_records = {}

        results = {}
        for path in paths:
            name = os.path.basename(path)
            results[name] = {}
            if re.match(ARBIN_CONFIG["file_pattern"], path):
                df = pd.read_csv(path, index_col=0)
                results[name]["validated"] = self.validate_arbin_dataframe(df)
                results[name]["method"] = self.validate_arbin_dataframe.__name__
            elif re.match(MACCOR_CONFIG["file_pattern"], path):
                df = pd.read_csv(path, delimiter="\t", skiprows=1)
                results[name]["validated"] = self.validate_maccor_dataframe(df)
                results[name]["method"] = self.validate_maccor_dataframe.__name__
            else:
                results[name]["validated"] = False
                results[name]["method"] = None
                self.errors = ["File type not recognized"]

            results[name]["time"] = json.dumps(
                datetime.now(), indent=4, sort_keys=True, default=str
            )
            results[name]["errors"] = self.errors

        if record_results:
            self.validation_records.update(results)
            dumpfn(self.validation_records, record_path)

        return results

    def _prevalidate_nonnull_column(self, df, column_name):
        """
        Scheme for prevalidation of non-numeric column,
        primarily used to pre-validate non-null cycle index,
        This is induces an error on validation of a dataframe
        with a null value in cycle_index, using a dummy doc
        and a custom schema for non-numeric cycle values.

        Args:
            df (pandas.DataFrame): dataframe.
            column_name: column identifier for pandas.

        Returns:
            bool: whether or not column contains a null value according to
                pandas.isnull().

        """
        null_index_mask = df[column_name].isnull()
        if null_index_mask.any():

            non_numeric_schema = {
                column_name: {"type": "list", "schema": {"type": "number"}}
            }
            dummy_df = df[null_index_mask].iloc[0:1].replace({np.nan: "None"})
            dummy_doc = dummy_df.to_dict(orient="list")
            self.validate(dummy_doc, schema=non_numeric_schema)
            return False
        else:
            return True


class SimpleValidator(object):
    """
    Lightweight class that does Dataframe-based, as
    opposed to dictionary based validation

    Note that schemas here are made to be identical to cerberus
    schemas and should support similar syntax, e. g.

    {COLUMN_NAME:
        {schema:
            type: TYPE_IN_COLUMN, [int, float, str, object]
            max: MAX_VALUE_IN_COLUMN,
            min: MIN_VALUE_IN_COLUMN
         type: list
        }
    }

    Note that the COLUMN_NAME.type key above is ignored, but
    COLUMN_NAME.schema.type is used.

    The only schema keys that are supported at this time are
    max, min, and type.

    Typing is compared using the key-mapping by rule defined
    by the ALLOWED_TYPES_BY_RULE attribute defined below.
    Supported type rules include "integer", "float", "numeric",
    and "string".  Note that type-checking for this class is
    not equivalent to checking types, and may involve custom
    logic which is defined in the check_type method below.
    """

    def __init__(self, schema_filename=DEFAULT_ARBIN_SCHEMA):
        """
        Args:
            schema_filename (str): filename corresponding to
                the schema
        """
        self.schema = loadfn(schema_filename)
        self.validation_records = None

    @staticmethod
    def check_type(df, type_rule):
        """
        Method to check type of input dataframe.

        Args:
            df (pandas.Dataframe): DataFrame.
            type_rule (str): string corresponding to type_rule
                to check, supported type rules are:
                integer: checks for numeric values which are
                    equal to their rounded values

        Returns:
            bool: valid
            str: verbose description of reason
        """
        if type_rule not in ["integer", "float", "numeric", "string"]:
            raise ValueError(
                "type_rule {} not supported, please choose one "
                "of integer, float, numeric, or string"
            )
        # Integer: Check residual from rounding
        if type_rule == "integer":
            nonint_indices = np.arange(len(df))[(df != np.round(df))]
            if nonint_indices.size > 0:
                value = df.iloc[nonint_indices[0]]
                return (
                    False,
                    "integer type check failed at index {} with value {}".format(
                        nonint_indices[0], value
                    ),
                )
        # Float: just check numpy dtyping
        elif type_rule == "float":
            if not np.issubdtype(df.dtype, np.floating):
                return False, "float type check failed, type is {}".format(df.dtype)

        # Numeric: check numpy number dtyping
        elif type_rule == "numeric":
            if not np.issubdtype(df.dtype, np.number):
                return False, "number type check failed, type is {}".format(df.dtype)

        # String: check string/unicode subdtype
        elif type_rule == "string":
            if not (
                np.issubdtype(df.dtype, np.object_)
                or np.issubdtype(df.dtype, np.unicode_)
            ):
                return False, "string type check failed, type is {}".format(df.dtype)
        return True, ""

    def validate(self, dataframe):
        """
        Method to run the validation on everything, and report
        the results, i. e. which columns are inconsistent with
        the schema.

        Args:
            dataframe (pandas.DataFrame): dataframe to be validated.

        Returns:
            dict: report corresponding to each validation
            str: reason for report validation failure, empty string on report
                validation success
        """
        dataframe = dataframe.rename(str.lower, axis="columns")
        for column_name, value in self.schema.items():
            column_schema = value["schema"]
            max_at_least_rule = column_schema.get("max_at_least")
            min_is_below_rule = column_schema.get("min_is_below")
            max_rule = column_schema.get("max")
            min_rule = column_schema.get("min")
            type_rule = column_schema.get("type")
            monotonic_rule = column_schema.get("monotonic")

            # Check type
            if type_rule is not None:
                validity, reason = self.check_type(
                    dataframe[column_name], type_rule=type_rule
                )
                if not validity:
                    reason = "Column {}: {}".format(column_name, reason)
                    return validity, reason

            # Check max
            if max_rule is not None:
                comp = np.where(dataframe[column_name] > max_rule)
                if comp[0].size > 0:
                    index = comp[0][0]
                    value = dataframe[column_name].iloc[index]
                    reason = (
                        "{} is higher than allowed max {} at index {}: "
                        "value={}".format(column_name, max_rule, index, value)
                    )
                    return False, reason

            # Check min
            if min_rule is not None:
                comp = np.where(dataframe[column_name] < min_rule)
                if comp[0].size > 0:
                    index = comp[0][0]
                    value = dataframe[column_name].iloc[index]
                    reason = (
                        "{} is lower than allowed min {} at index {}:"
                        "value={}".format(column_name, min_rule, index, value)
                    )
                    return False, reason

            # Check a maximum value is at least above a threshold
            if max_at_least_rule is not None:
                comp = np.where(dataframe[column_name].max() < max_at_least_rule)
                if comp[0].size > 0:
                    index = comp[0][0]
                    value = dataframe[column_name].iloc[index]
                    reason = (
                        "{} needs to reach at least {} for processing, instead found:"
                        "value={}".format(column_name, max_at_least_rule, value)
                    )
                    return False, reason

            # Check a minimum value is below above a threshold
            if min_is_below_rule is not None:
                comp = np.where(dataframe[column_name].min() > min_is_below_rule)
                if comp[0].size > 0:
                    index = comp[0][0]
                    value = dataframe[column_name].iloc[index]
                    reason = (
                        "{} needs to reach under {} for processing, instead found:"
                        "value={}".format(column_name, max_at_least_rule, value)
                    )
                    return False, reason

            if monotonic_rule == 'increasing':
                diff_series = dataframe[column_name].diff().dropna()
                if len(diff_series[diff_series < 0]) > 0:
                    reason = (
                        "{} needs to be monotonically increasing for processing".format(column_name)
                    )
                    return False, reason

        return True, ""

    def validate_from_paths(
        self,
        paths,
        record_results=False,
        skip_existing=False,
        record_path=DEFAULT_VALIDATION_RECORDS,
    ):
        """
        This method streamlines validation of multiple Arbin csv files given a list of paths.

        It can also do bookkeeping of validations by dumping results in a json file,
        locally until a more centralized method is implemented.

        Args:
            paths (list): a list of paths to csv files
            record_results (bool): Whether to record the validation results locally or not (defaults to False)
            skip_existing (bool): Whether to skip already validated files. This is done by checking if the
                                    file is in the validation_records. skip_existing only matters if record_results
                                    is True. (defaults to False)
            record_path (str): path to the json file storing the past validation results.
        Returns:
            dict: Results of the validation in the form of a key,value pairs where each key corresponds to the filename
                validated. For each file, the results contain a field "validated", True if validation was successful or
                False if not. "errors", "method" and "time" are simply the errors encountered during validation, method
                used for validation, and time of validation, respectively.

        """
        if record_results:
            if os.path.isfile(record_path):
                self.validation_records = loadfn(record_path)
                if skip_existing:
                    paths = [
                        path
                        for path in paths
                        if os.path.basename(path) not in self.validation_records
                    ]
            else:
                self.validation_records = {}

        results = {}
        for path in tqdm(paths):
            name = os.path.basename(path)
            results[name] = {}
            project_schema = loadfn(PROJECT_SCHEMA)
            if re.match(ARBIN_CONFIG["file_pattern"], path):
                if project_schema.get(name.split("_")[0], {}).get("arbin"):
                    schema_filename = os.path.join(
                        VALIDATION_SCHEMA_DIR, project_schema.get(name.split("_")[0], {}).get("arbin")
                    )
                    self.schema = loadfn(schema_filename)
                    method = project_schema.get(name.split("_")[0], {}).get("arbin")
                else:
                    schema_filename = os.path.join(
                        VALIDATION_SCHEMA_DIR, "schema-arbin-lfp.yaml"
                    )
                    self.schema = loadfn(schema_filename)
                    method = "simple_arbin"

                df = pd.read_csv(path, index_col=0)
                validated, reason = self.validate(df)
            elif re.match(MACCOR_CONFIG["file_pattern"], path):
                if project_schema.get(name.split("_")[0], {}).get("maccor"):
                    schema_filename = os.path.join(
                        VALIDATION_SCHEMA_DIR, project_schema.get(name.split("_")[0], {}).get("maccor")
                    )
                    self.schema = loadfn(schema_filename)
                    method = project_schema.get(name.split("_")[0], {}).get("maccor")
                else:
                    schema_filename = os.path.join(
                        VALIDATION_SCHEMA_DIR, "schema-maccor-2170.yaml"
                    )
                    self.schema = loadfn(schema_filename)
                    method = "simple_maccor"
                self.allow_unknown = True
                df = pd.read_csv(path, delimiter="\t", skiprows=1)

                # Columns need to be retyped and renamed for validation,
                # conversion will happen during structuring
                df["State"] = df["State"].astype(str)
                df["current"] = df["Amps"]

                validated, reason = self.validate(df)

            else:
                validated, reason = False, "File type not recognized"
                method = None
            results[name].update(
                {
                    "validated": validated,
                    "method": method,
                    "errors": reason,
                    "time": json.dumps(
                        datetime.now(), indent=4, sort_keys=True, default=str
                    ),
                }
            )

            if validated:
                logger.info("%s method=%s errors=%s", name, method, reason, extra=s)
            else:
                logger.warning("%s method=%s errors=%s", name, method, reason, extra=s)

        if record_results:
            self.validation_records.update(results)
            dumpfn(self.validation_records, record_path)

        return results


class BeepValidationError(Exception):
    """Custom error to raise when validation fails"""


def validate_file_list_from_json(
    file_list_json,
    record_results=False,
    skip_existing=False,
    validator_class=SimpleValidator,
):
    """
    Validates a list of files from json input

    Args:
        file_list_json (str): input for validation files, should be a json string
            with attribute "file_list" or a filename (e. g. something.json)
            corresponding to a json object with a similar attribute.
        record_results (bool): Whether to record the validation results locally
            or not (defaults to False).
        skip_existing (bool): Whether to skip already validated files. This
            is done by checking if the file is in the validation_records.
            skip_existing only matters if record_results is True. (defaults to False)
        validator_class (ValidatorBeep or SimpleValidator): validator class
            to use in validation.

    Returns:
        str: json dump of the validator results.

    """
    # Process input json
    if file_list_json.endswith(".json"):
        file_list_data = loadfn(file_list_json)
    else:
        file_list_data = json.loads(file_list_json)

    # Setup workflow
    outputs = WorkflowOutputs()

    file_list = file_list_data["file_list"]

    validator = validator_class()
    all_results = validator.validate_from_paths(
        file_list, record_results=record_results, skip_existing=skip_existing,
    )

    # Get validities and recast to strings (valid/invalid) based on result
    validity = [all_results[os.path.split(file)[-1]]["validated"] for file in file_list]

    validity = list(map(lambda x: "valid" if x else "invalid", validity))

    # Get errors
    errors = [all_results[os.path.split(file)[-1]]["errors"] for file in file_list]
    messages = [{"comment": "", "error": error} for error in errors]
    output_json = {
        "file_list": file_list,
        "run_list": file_list_data["run_list"],
        "validity": validity,
        "message_list": messages,
    }

    # Workflow outputs
    file_list_size = len(output_json["file_list"])
    if file_list_size > 1 or file_list_size == 0:
        logger.warning("{file_list_size} files being validated, should be 1")

    output_data = {
        "filename": output_json["file_list"][0],
        "run_id": output_json["run_list"][0],
        "result": output_json["validity"][0],
    }

    outputs.put_workflow_outputs(output_data, "validating")

    return json.dumps(output_json)


def main():
    """
    Main function for running of this module as a script

    Returns:
        (None)

    """
    logger.info("starting", extra=s)
    logger.info("Running version=%s", __version__, extra=s)
    try:
        args = docopt(__doc__)
        input_json = args["INPUT_JSON"]
        print(validate_file_list_from_json(input_json), end="")
    except Exception as e:
        logger.error(str(e), extra=s)
        raise e
    logger.info("finish", extra=s)
    return None


if __name__ == "__main__":
    main()



#  Copyright (c) 2019 Toyota Research Institute
"""
Top-level module for beep.  Put anything that should
be available in the beep namespace here.
"""
import os
import logging
import sys
from functools import partial
import numpy as np
import watchtower
import time
from botocore.exceptions import NoCredentialsError
from tqdm import tqdm as _tqdm

from .config import config

try:
    from io import StringIO
except (ImportError, ModuleNotFoundError):
    from StringIO import StringIO

# Versioning.  The python code version is frequently tagged
# with a commit hash from the repo, which is supplied via
# an environment variable by the integration build procedure
__version__ = "2021.9.30.15"
VERSION_TAG = os.environ.get("BEEP_VERSION_TAG")
if VERSION_TAG is not None:
    __version__ = "-".join([__version__, VERSION_TAG])

# Custom tqdm with optional turnoff from env
tqdm = partial(_tqdm, disable=bool(os.environ.get("TQDM_OFF")))

ENV_VAR = "BEEP_ENV"
PROCESSED_DIR = "BEEP_PROCESSING_DIR"
MAX_RETRIES = 12

# environment
ENVIRONMENT = os.getenv(ENV_VAR)
if ENVIRONMENT is None or ENVIRONMENT not in config.keys():
    ENVIRONMENT = "local"

DIR = os.getenv(PROCESSED_DIR)
if DIR is None:
    if ENVIRONMENT in ["stage", "prod"]:
        os.environ[PROCESSED_DIR] = "/"
    elif ENVIRONMENT in ["local", "dev", "test"]:
        os.environ[PROCESSED_DIR] = os.getcwd()
    else:
        raise ValueError(
            f"The directory for processing cycling data {PROCESSED_DIR} must be set"
            + f" eg. /Users/Bob/cycling"
        )

MODULE_DIR = os.path.dirname(__file__)
CONVERSION_SCHEMA_DIR = os.path.join(MODULE_DIR, "conversion_schemas")
VALIDATION_SCHEMA_DIR = os.path.join(MODULE_DIR, "validation_schemas")
MODEL_DIR = os.path.join(MODULE_DIR, "models")

LOG_DIR = os.path.join(MODULE_DIR, "logs")
if not os.path.isdir(LOG_DIR):
    os.mkdir(LOG_DIR)

# Get S3 cache location from env or use default in repo
S3_CACHE = os.environ.get("BEEP_S3_CACHE", os.path.join(MODULE_DIR, "..", "s3_cache"))

# logging
np.set_printoptions(precision=3)

# service (logging)
container = config[ENVIRONMENT]["logging"]["container"]

# initialize logger and clear previous handlers and filters, if exist
logger = logging.getLogger(ENVIRONMENT + "/beep")
logger.handlers = []
logger.filters = []

fmt_str = (
    '{"time": "%(asctime)s", "level": "%(levelname)s", '
    '"service": "%(service)s", "process": "%(process)d", '
    '"module": "%(module)s", "func": "%(funcName)s", '
    '"msg": "%(message)s"}'
)
formatter = logging.Formatter(fmt_str)

# output and format
if "CloudWatch" in config[ENVIRONMENT]["logging"]["streams"]:
    if ENVIRONMENT == "stage":
        for _ in range(MAX_RETRIES):
            try:
                hdlr = watchtower.CloudWatchLogHandler(log_group="/stage/beep/services")
            except NoCredentialsError:
                time.sleep(10)
                continue
            else:
                break
        else:
            raise NoCredentialsError
    else:
        hdlr = watchtower.CloudWatchLogHandler(log_group="Worker")
    hdlr.setFormatter(formatter)
    logger.addHandler(hdlr)
if "stdout" in config[ENVIRONMENT]["logging"]["streams"]:
    hdlr = logging.StreamHandler(sys.stdout)
    hdlr.setFormatter(formatter)
    logger.addHandler(hdlr)
if "file" in config[ENVIRONMENT]["logging"]["streams"]:
    log_file = os.path.join(MODULE_DIR, "Testing_logger.log")
    hdlr = logging.FileHandler(log_file, "a")
    hdlr.setFormatter(formatter)
    logger.addHandler(hdlr)

logger.setLevel("DEBUG")
logger.propagate = False



# Copyright [2020] [Toyota Research Institute]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Module for Arbin-compatible schedule file
parsing and parameter insertion
"""
import re
import warnings
from copy import deepcopy
from collections import OrderedDict
from beep.utils import DashOrderedDict


class Schedule(DashOrderedDict):
    """
    Schedule file utility. Provides the ability to read
    an Arbin type schedule file.  Note that __init__ works
    identically to that of an OrderedDict, i. e. with
    tuple or dictionary inputs and inherits from pydash
    getters and setters with Schedule.get, Schedule.set
    and Schedule.unset e.g.

    >>> schedule = Schedule.from_file("arbin_file_1.sdu")
    >>> schedule.set("Schedule.Step7.m_szLabel", "CC1")
    >>> print(schedule['Schedule']['Step7']['m_szLabel'])
    >>> "CC1"

    """

    @classmethod
    def from_file(cls, filename, encoding="latin-1"):
        """
        Schedule file ingestion. Converts a schedule file with section headers
        to an ordered dict with section headers as nested dicts.

        Args:
            filename (str): Schedule file name (tested with FastCharge schedule file)
            encoding (str): encoding of schedule file

        Returns:
            (Schedule): Ordered dictionary with keys corresponding to options
                or control variables. Section headers are nested dicts within
                the dict

        """
        obj = cls()
        with open(filename, "rb") as f:
            text = f.read()
            text = text.decode(encoding)

        split_text = re.split(r"\[(.+)\]", text)
        for heading, body in zip(split_text[1::2], split_text[2::2]):
            body_lines = re.split(r"[\r\n]+", body.strip())
            body_dict = OrderedDict([line.split("=", 1) for line in body_lines])
            heading = heading.replace("_", ".")
            obj.set(heading, body_dict)

        return obj

    def to_file(self, filename, encoding="latin-1", linesep="\r\n"):
        """
        Schedule file output. Converts an dictionary to a schedule file with
        the appropriate section headers. This function
        DOES NOT check the flow control or limits set in the steps. The dictionary
        must represent a valid schedule before it is passed to this function.

        Args:
            filename (str): string corresponding to the file to
                output the schedule to
            encoding (str): text encoding for the file
            linesep (str): line separator for the file,
                default Windows-compatible "\r\n"

        """
        # Flatten dict
        data = deepcopy(self)
        flat_keys = _get_headings(data, delimiter=".")
        flat_keys.reverse()  # Reverse ensures sub-dicts are removed first
        data_tuples = []
        for flat_key in flat_keys:
            data_tuple = (flat_key.replace(".", "_"), data.get_path(flat_key))
            data_tuples.append(data_tuple)
            data.unset(flat_key)
        data_tuples.reverse()

        # Construct text
        blocks = []
        for section_title, body_data in data_tuples:
            section_header = "[{}]".format(section_title)
            body = linesep.join(
                ["=".join([key, value]) for key, value in body_data.items()]
            )
            blocks.append(linesep.join([section_header, body]))
        contents = linesep.join(blocks) + linesep

        # Write file
        with open(filename, "wb") as f:
            f.write(contents.encode(encoding))

    @classmethod
    def from_fast_charge(cls, CC1, CC1_capacity, CC2, template_filename):
        """
        Function takes parameters for the FastCharge Project
        and creates the schedule files necessary to run each of
        these parameter combinations. Assumes that control type
        is CCCV.

        Args:
            CC1 (float): Constant current value for charge section 1 in Amps
            CC1_capacity (float): Capacity to charge to for section 1 in Amp-hours
            CC2 (float): Constant current value for charge section 2 in Amps
            template_filename (str): File path to pull the template schedule
                file from

        """
        obj = cls.from_file(template_filename)

        obj.set_labelled_steps(
            "CC1", "m_szCtrlValue", step_value="{0:.3f}".format(CC1).rstrip("0")
        )
        obj.set_labelled_limits(
            "CC1",
            "PV_CHAN_Charge_Capacity",
            comparator=">",
            value="{0:.3f}".format(CC1_capacity).rstrip("0"),
        )
        obj.set_labelled_steps(
            "CC2", "m_szCtrlValue", step_value="{0:.3f}".format(CC2).rstrip("0")
        )
        return obj

    def get_labelled_steps(self, step_label):
        """
        Insert values for steps in the schedule section

        Args:
            step_label (str): The user determined step label for the step.
                If there are multiple identical labels this will operate
                on the first one it encounters

        Returns:
            (iterator): iterator for subkeys of schedule which match
                the label value
        """
        # Find all step labels
        labelled_steps = filter(
            lambda x: self.get_path("Schedule.{}.m_szLabel".format(x)) == step_label,
            self["Schedule"].keys(),
        )
        return labelled_steps

    def set_labelled_steps(self, step_label, step_key, step_value, mode="first"):
        """
        Insert values for steps in the schedule section

        Args:
            step_label (str): The user determined step label for the step.
                If there are multiple identical labels this will operate
                on the first one it encounters
            step_key (str): Key in the step to set, e.g. ('m_szStepCtrlType')
            step_value (str): Value to set for the key
            mode (str): accepts 'first' or 'all',
                for 'first' updates only first step with matching label
                for 'all' updates all steps with matching labels

        Returns:
            dict: Altered ordered dictionary with keys corresponding to
                options or control variables.
        """
        # Find all step labels
        labelled_steps = self.get_labelled_steps(step_label)

        # TODO: should update happen in place or return new?
        for step in labelled_steps:
            self.set("Schedule.{}.{}".format(step, step_key), step_value)
            if mode == "first":
                break

        return self

    def set_labelled_limits(self, step_label, limit_var, comparator, value):
        """
        Insert values for the limits in the steps in the schedule section

        Args:
            step_label (str): The user determined step label for the step. If there
                are multiple identical labels this will operate on the first one it
                encounters
            limit_var (str): Variable being used for this particular limit in the step
            value (int or str): threshold value to trip limit
            comparator (str): str-represented comparator to trip limit,
                e.g. '>' or '<'

        Returns:
            dict: Altered ordered dictionary with keys corresponding to options or control
                variables.
        """
        labelled_steps = self.get_labelled_steps(step_label)
        for step in labelled_steps:
            # Get all matching limit keys
            step_data = self.get_path("Schedule.{}".format(step))
            limits = [
                heading
                for heading in _get_headings(step_data)
                if heading.startswith("Limit")
            ]
            # Set limit of first limit step with matching code
            for limit in limits:
                limit_data = step_data[limit]
                if limit_data["m_bStepLimit"] == "1":  # Code corresponding to stop
                    if limit_data["Equation0_szLeft"] == limit_var:
                        limit_prefix = "Schedule.{}.{}".format(step, limit)
                        self.set(
                            "{}.Equation0_szCompareSign".format(limit_prefix),
                            comparator,
                        )
                        self.set("{}.Equation0_szRight".format(limit_prefix), value)
                    else:
                        warnings.warn(
                            "Additional step limit at {}.{}".format(step, limit)
                        )
        return self


def _get_headings(obj, delimiter="."):
    """
    Utility function for getting all nested keys
    of a dictionary whose values are themselves
    a dict

    Args:
        obj (dict): nested dictionary to be searched
        delimiter (str): string delimiter for nested
            sub_headings, e. g. top_middle_low for
            'top', 'middle', and 'low' nested keys

    """
    headings = []
    for heading, body in obj.items():
        if isinstance(body, dict):
            headings.append(heading)
            sub_headings = _get_headings(body, delimiter=delimiter)
            headings.extend(
                [delimiter.join([heading, sub_heading]) for sub_heading in sub_headings]
            )
    return headings



"""
Module for S3 operations
"""
import boto3


def download_s3_object(bucket, key, destination_path):
    """
    Download an object from a bucket/key location on S3 to local disk.

    Args:
        bucket (str):                   S3 bucket
        key (str):                      S3 key
        destination_path (str):         local path at which file object is to be stored
    """
    s3_client = boto3.client('s3')
    s3_client.download_file(bucket, key, destination_path)
    return



import pandas as pd

from beep.protocol.maccor import Procedure

parameters = pd.read_csv("./parameters.csv")
for index, protocol_params in parameters.iterrows():
    proc = Procedure.from_file("./testV1.000")
    proc["MaccorTestProcedure"]["ProcSteps"]["TestStep"][8]["StepValue"] = protocol_params["power"]
    proc.to_file("./project{}.000".format(str(index)))



import os
import pandas as pd
from monty.serialization import loadfn

from beep.features.base import BeepFeatures
from beep.features.intracell_analysis import IntracellAnalysis
from beep import MODULE_DIR


FEATURE_HYPERPARAMS = loadfn(
    os.path.join(MODULE_DIR, "features/feature_hyperparameters.yaml")
)

s = {"service": "DataAnalyzer"}


class IntracellCycles(BeepFeatures):
    """
    Object corresponding to the fitted material parameters of the cell. Material parameters
    are determined by using high resolution half cell data to fit full cell dQdV curves. Rows
    of the output dataframe correspond to each of the diagnostics throughout the life of the
    cell.
        name (str): predictor object name.
        X (pandas.DataFrame): features in DataFrame format.
        metadata (dict): information about the conditions, data
            and code used to produce features
    """

    # Class name for the feature object
    class_feature_name = "IntracellCycles"

    def __init__(self, name, X, metadata):
        """
        Args:
            name (str): predictor object name
            X (pandas.DataFrame): features in DataFrame format.
            metadata (dict): information about the data and code used to produce features
        """
        super().__init__(name, X, metadata)
        self.name = name
        self.X = X
        self.metadata = metadata

    @classmethod
    def validate_data(cls, processed_cycler_run, params_dict=None):
        """
        This function determines if the input data has the necessary attributes for
        creation of this feature class. It should test for all of the possible reasons
        that feature generation would fail for this particular input data.

        Args:
            processed_cycler_run (beep.structure.ProcessedCyclerRun): data from cycler run
            params_dict (dict): dictionary of parameters governing how the ProcessedCyclerRun object
                gets featurized. These could be filters for column or row operations

        Returns:
            bool: True/False indication of ability to proceed with feature generation
        """
        conditions = []

        if not hasattr(processed_cycler_run, "diagnostic_summary") & hasattr(
            processed_cycler_run, "diagnostic_data"
        ):
            return False
        if processed_cycler_run.diagnostic_summary is None:
            return False
        elif processed_cycler_run.diagnostic_summary.empty:
            return False
        else:
            # Ensure overlap of cycle indices above threshold and matching cycle type
            eol_cycle_index_list = processed_cycler_run.diagnostic_summary[
                (processed_cycler_run.diagnostic_summary.cycle_type == params_dict["diagnostic_cycle_type"]) &
                (processed_cycler_run.diagnostic_summary.discharge_capacity > IntracellAnalysis.THRESHOLD)
                ].cycle_index.to_list()
            if not eol_cycle_index_list:
                return False

            conditions.append(
                any(
                    [
                        "rpt" in x
                        for x in processed_cycler_run.diagnostic_summary.cycle_type.unique()
                    ]
                )
            )

        return all(conditions)

    @classmethod
    def features_from_processed_cycler_run(cls, processed_cycler_run, params_dict=None,
                                           parameters_path="data-share/raw/parameters",
                                           cell_info_path="data-share/raw/cell_info"):
        """
        Args:
            processed_cycler_run (beep.structure.ProcessedCyclerRun)
            params_dict (dict): dictionary of parameters governing how the ProcessedCyclerRun object
                gets featurized. These could be filters for column or row operations
            parameters_path (str): Root directory storing project parameter files.
            cell_info_path (str): Root directory for cell half cell data

        Returns:
             (pd.DataFrame) containing the cell material parameters as a function of cycle index
        """
        if params_dict is None:
            params_dict = FEATURE_HYPERPARAMS[cls.class_feature_name]

        cell_dir = os.path.join(
            os.environ.get("BEEP_PROCESSING_DIR", "/"), cell_info_path
        )
        ia = IntracellAnalysis(os.path.join(cell_dir, params_dict["cathode_file"]),
                               os.path.join(cell_dir, params_dict["anode_file"]),
                               cycle_type=params_dict["diagnostic_cycle_type"],
                               step_type=params_dict["step_type"]
                               )

        (cell_init_aligned, cell_init_profile, PE_matched, NE_matched) = ia.intracell_wrapper_init(
            processed_cycler_run,
        )

        eol_cycle_index_list = processed_cycler_run.diagnostic_summary[
            (processed_cycler_run.diagnostic_summary.cycle_type == ia.cycle_type) &
            (processed_cycler_run.diagnostic_summary.discharge_capacity > ia.THRESHOLD)
            ].cycle_index.to_list()

        # initialize dicts before for loop
        dataset_dict_of_cell_degradation_path = dict()
        real_cell_dict_of_profiles = dict()
        for i, cycle_index in enumerate(eol_cycle_index_list):
            loss_dict, profiles_dict = ia.intracell_values_wrapper(cycle_index,
                                                                   processed_cycler_run,
                                                                   cell_init_aligned,
                                                                   cell_init_profile,
                                                                   PE_matched,
                                                                   NE_matched,
                                                                   )
            dataset_dict_of_cell_degradation_path.update(loss_dict)
            real_cell_dict_of_profiles.update(profiles_dict)

        degradation_df = pd.DataFrame(dataset_dict_of_cell_degradation_path,
                                      index=['LLI', 'LAM_PE', 'LAM_NE', 'x_NE_2', 'alpha_real', 'alpha_emulated',
                                             'PE_upper_voltage', 'PE_lower_voltage', 'PE_upper_SOC', 'PE_lower_SOC',
                                             'PE_mass', 'NE_upper_voltage', 'NE_lower_voltage', 'NE_upper_SOC',
                                             'NE_lower_SOC', 'NE_mass', 'Li_mass'
                                             ]).T
        return degradation_df


class IntracellFeatures(IntracellCycles):
    """
    Object corresponding to the fitted material parameters of the cell. Material parameters
    are determined by using high resolution half cell data to fit full cell dQdV curves. The
    parameters from the first and second diagnostics are used as the feature values.
        name (str): predictor object name.
        X (pandas.DataFrame): features in DataFrame format.
        metadata (dict): information about the conditions, data
            and code used to produce features
    """

    # Class name for the feature object
    class_feature_name = "IntracellFeatures"

    @classmethod
    def features_from_processed_cycler_run(cls, processed_cycler_run, params_dict=None,
                                           parameters_path="data-share/raw/parameters",
                                           cell_info_path="data-share/raw/cell_info"):
        """
        Args:
            processed_cycler_run (beep.structure.ProcessedCyclerRun)
            params_dict (dict): dictionary of parameters governing how the ProcessedCyclerRun object
                gets featurized. These could be filters for column or row operations
            parameters_path (str): Root directory storing project parameter files.
            cell_info_path (str): Root directory for cell half cell data


        Returns:
             (pd.DataFrame) containing the cell material parameters for the first and second diagnotics
                as a single row dataframe
        """
        if params_dict is None:
            params_dict = FEATURE_HYPERPARAMS[cls.class_feature_name]

        cell_dir = os.path.join(
            os.environ.get("BEEP_PROCESSING_DIR", "/"), cell_info_path
        )
        ia = IntracellAnalysis(os.path.join(cell_dir, params_dict["cathode_file"]),
                               os.path.join(cell_dir, params_dict["anode_file"]),
                               cycle_type=params_dict["diagnostic_cycle_type"],
                               step_type=params_dict["step_type"]
                               )

        (cell_init_aligned, cell_init_profile, PE_matched, NE_matched) = ia.intracell_wrapper_init(
            processed_cycler_run,
        )

        eol_cycle_index_list = processed_cycler_run.diagnostic_summary[
            (processed_cycler_run.diagnostic_summary.cycle_type == ia.cycle_type) &
            (processed_cycler_run.diagnostic_summary.discharge_capacity > ia.THRESHOLD)
            ].cycle_index.to_list()

        # # initializations before for loop
        dataset_dict_of_cell_degradation_path = dict()
        real_cell_dict_of_profiles = dict()
        for i, cycle_index in enumerate(eol_cycle_index_list[0:2]):
            loss_dict, profiles_dict = ia.intracell_values_wrapper(cycle_index,
                                                                   processed_cycler_run,
                                                                   cell_init_aligned,
                                                                   cell_init_profile,
                                                                   PE_matched,
                                                                   NE_matched,
                                                                   )
            dataset_dict_of_cell_degradation_path.update(loss_dict)
            real_cell_dict_of_profiles.update(profiles_dict)

        degradation_df = pd.DataFrame(dataset_dict_of_cell_degradation_path,
                                      index=['LLI', 'LAM_PE', 'LAM_NE', 'x_NE_2', 'alpha_real', 'alpha_emulated',
                                             'PE_upper_voltage', 'PE_lower_voltage', 'PE_upper_SOC', 'PE_lower_SOC',
                                             'PE_mass', 'NE_upper_voltage', 'NE_lower_voltage', 'NE_upper_SOC',
                                             'NE_lower_SOC', 'NE_mass', 'Li_mass'
                                             ]).T
        diag_0_names = ["diag_0_" + name for name in degradation_df.columns]
        diag_1_names = ["diag_1_" + name for name in degradation_df.columns]
        values = {0: degradation_df.iloc[0].tolist() + degradation_df.iloc[1].tolist()}
        features_df = pd.DataFrame(values, index=diag_0_names+diag_1_names).T
        return features_df



"""Base classes for EIS.

Use these classes to create new EIS-capable BEEPDatapath classes for EIS-capable cyclers.

"""
import abc

import pandas as pd
from monty.json import MSONable

from beep.structure.base import BEEPDatapath


class EIS(MSONable):
    """Base class for an Electrochemical Impedance Spectrum

    EIS can be used by itself to hold EIS data; alternatively, it can be
    inherited by child *EIS classes (e.g., MaccorEIS) for specific cyclers
    which have their own ingestion or anaylsis formats.

    Each EIS instance corresponds to a single electrochemical impedance spectrum.

    This class does not need to implement a from_file method, though
    if you intend for it to be used with BEEPDatapathWithEIS, it should
    implement from_file.

    Attributes:
        data (pd.DataFrame): The EIS data, as a dataframe
        metadata (pd.DataFrame): The EIS metadata, as a dataframe
    """

    def __init__(self, data, metadata):
        """
        Args:
            data (pd.DataFrame): The EIS data, as a dataframe
            metadata (pd.DataFrame): The EIS metadata, as a dataframe
        """
        self.data = data
        self.metadata = metadata

    @classmethod
    def from_file(cls, filename):
        """Create an EIS object from a raw data file.

        Args:
            filename (str, Pathlike): The filename of the raw EIS file.

        Returns:

        """
        raise NotImplementedError

    def as_dict(self):
        """Method for serialization as dictionary

        Returns:
            ({}): dictionary representation of this object

        """
        return {
            "@module": self.__class__.__module__,
            "@class": self.__class__.__name__,
            "data": self.data.to_dict("list"),
            "metadata": self.metadata.to_dict()
        }

    @classmethod
    def from_dict(cls, d):
        """Method of invocation from dictionary

        Args:
            d ({}): dictionary from which to invoke

        Returns:
            (EISpectrum): object invoked from dictionary

        """
        data = pd.DataFrame(d["data"])
        data = data.sort_index()
        metadata = pd.DataFrame(d["metadata"])
        return cls(data, metadata)


class BEEPDatapathWithEIS(BEEPDatapath):
    """An ABC for BEEPDatapaths capable of handling EIS data.

    BEEPDatapathWithEIS can be inherited by child cycler classes for handling and structuring
    cycler runs with EIS data. This does NOT mean that BEEPDatapathWithEIS must contain EIS data.
    All the capability of normal BEEPDatapathWithEIS is retained; extra features are added.

    All BEEPDatapathWithEIS MUST implement a load_eis method.

    Attributes:
        - eis ([EIS]): A list of EIS or EIS child objects; i.e., 0 or more electrochemical impedance spectra.
    """

    def __init__(self, *args, **kwargs):
        self.eis = None
        super(BEEPDatapathWithEIS, self).__init__(*args, **kwargs)

    @abc.abstractmethod
    def load_eis(self, *args, **kwargs):
        """Load one or more EIS files.

        Implement this method to load one or more EIS files.

        load_eis should do both of the following:
            - add all of the eis paths to the BEEPDatapathWithEIS.paths attribute like:
                self.paths["eis"] = <<list_of_eis_paths>>

            - update BEEPDatapathWithEIS.eis as a list of EIS or EIS-derived objects:
                self.eis = [EIS_obj1, EIS_obj2, EIS_obj3, ....]


        Args:
            *args:
            **kwargs:

        Returns:
            None
        """
        raise NotImplementedError(
            "EIS containing datapath must implement 'load_eis' method."
        )



"""
Module for computing waveforms
"""
import pandas as pd
import numpy as np
from scipy import interpolate
from scipy.optimize import fsolve


def convert_velocity_to_power_waveform(waveform_file, velocity_units):
    """
    Helper function to perform model based conversion of velocity waveform into power waveform.

    For model description and parameters ref JECS, 161 (14) A2099-A2108 (2014)
    "Model-Based SEI Layer Growth and Capacity Fade Analysis for EV and PHEV Batteries and Drive Cycles"

    Args:
        waveform_file (str): file containing tab or comma delimited values of time and velocity
        velocity_units (str): units of velocity. Accept 'mph' or 'kmph' or 'mps'

    returns
    pd.DataFrame with two columns: time (sec) and power (W). Negative = Discharge
    """
    df = pd.read_csv(waveform_file, sep="\t", header=0)
    df.columns = ["t", "v"]

    if velocity_units == "mph":
        scale = 1600.0 / 3600.0
    elif velocity_units == "kmph":
        scale = 1000.0 / 3600.0
    elif velocity_units == "mps":
        scale = 1.0
    else:
        raise NotImplementedError

    df.v = df.v * scale

    # Define model constants
    m = 1500  # kg
    rolling_resistance_coef = 0.01  # rolling resistance coeff
    g = 9.8  # m/s^2
    theta = 0  # gradient in radians
    rho = 1.225  # kg/m^3
    drag_coef = 0.34  # Coeff of drag
    frontal_area = 1.75  # m^2
    v_wind = 0  # wind velocity in m/s

    # Power = Force * vel
    # Force = Rate of change of momentum + Rolling frictional force + Aerodynamic drag force

    # Method treats the time-series as is and does not interpolate on a uniform grid before computing gradient.
    power = (
        m * np.gradient(df.v, df.t)
        + rolling_resistance_coef * m * g * np.cos(theta * np.pi / 180)
        + 0.5 * rho * drag_coef * frontal_area * (df.v - v_wind) ** 2
    )

    power = -power * df.v  # positive power = charge

    return pd.DataFrame({"time": df.t, "power": power})


class RapidChargeWave:
    """
       Object to produce charging waveforms, with direct comparison between stepwise and
       smooth charging waveforms. Algorithms written by Patrick Asinger.

    """
    def __init__(self,
                 final_c_rate,
                 soc_initial,
                 soc_final,
                 max_c_rate,
                 min_c_rate
                 ):
        """
        Args:
            final_c_rate (float): charging rate for the final step
            soc_initial (float): estimated starting soc for the fast charging portion of the cycle
            soc_final (float): estimated soc to end the fast charging portion of the cycle
            max_c_rate (float): maximum charging rate for the charge
            min_c_rate (float): minimum charging rate for the charge

        """
        self.final_c_rate = final_c_rate
        self.soc_i = soc_initial
        self.soc_f = soc_final
        self.max_c_rate = max_c_rate
        self.min_c_rate = min_c_rate
        self.soc_points = 1000

    def get_currents_with_uniform_time_basis(self, charging_c_rates, mesh_points):
        """
        Function to re-interpolate all of the current values to a uniform time basis, with each time step 1 sec

        Args:
            charging_c_rates (list): c-rates for each of the charging steps. Each step is assumed to be an equal
                SOC portion of the charge, and the length of the list just needs to be at least 1
            mesh_points (list): soc values for beginning and end of each of the charging windows

        returns
        np.array: array with the adjusted smooth current as a function of uniform time basis
        np.array: array with the multistep current as a function of the uniform time basis
        np.array: array with the corresponding uniformly spaced time values
        """
        assert np.all(np.diff(mesh_points)) > 0
        current_smooth, time_smooth, current_multistep, time_multistep = \
            self.get_input_currents_both_to_final_soc(charging_c_rates, mesh_points)
        end_time_smooth = int(np.round(np.max(time_smooth), 0))
        end_time_multistep = int(np.round(np.max(time_multistep), 0))
        assert end_time_smooth == end_time_multistep

        time_uniform = np.arange(0, end_time_smooth, 1)
        smooth_fill = (current_smooth[0], current_smooth[-1])

        interpolate_smooth = interpolate.interp1d(time_smooth,
                                                  current_smooth,
                                                  bounds_error=False,
                                                  fill_value=smooth_fill)
        current_smooth_uniform = interpolate_smooth(time_uniform)

        multistep_fill = (current_multistep[0], current_multistep[-1])
        interpolate_multistep = interpolate.interp1d(time_multistep,
                                                     current_multistep,
                                                     bounds_error=False,
                                                     fill_value=multistep_fill
                                                     )
        current_multistep_uniform = interpolate_multistep(time_uniform)

        return current_smooth_uniform, current_multistep_uniform, time_uniform

    def get_input_currents_both_to_final_soc(self, charging_c_rates, mesh_points):
        """
        Function to shift the charging rates for the smooth current to ensure that both the multistep and
        smooth charging functions reach the end in the same length of time.

        Args:
            charging_c_rates (list): c-rates for each of the charging steps. Each step is assumed to be an equal
                SOC portion of the charge, and the length of the list just needs to be at least 1
            mesh_points (list): soc values for beginning and end of each of the charging windows

        returns
        np.array: array with the adjusted smooth current as a function of soc
        np.array: array with the corresponding time values
        np.array: array with the multistep current as a function of soc
        np.array: array with the corresponding time values
        """

        current_multistep, soc_vector = self.get_input_current_multistep_soc_as_x(charging_c_rates, mesh_points)

        time_multistep = self.get_time_vector_from_c_vs_soc(soc_vector, current_multistep)
        end_soc_time = time_multistep[soc_vector >= self.soc_f][0]

        offset_solved = fsolve(self.offset_value,
                               0.01,
                               maxfev=1000,
                               factor=0.1,
                               epsfcn=0.01,
                               args=(end_soc_time, charging_c_rates, mesh_points))

        charging_c_rates = [charging_c_rates[0]] + list(np.add(charging_c_rates[1:-1], offset_solved)) + \
                           [charging_c_rates[-1]]

        current_smooth_time_adjusted, soc_vector = self.get_input_current_smooth_soc_as_x(charging_c_rates, mesh_points)
        time_smooth_time_adjusted = self.get_time_vector_from_c_vs_soc(soc_vector, current_smooth_time_adjusted)

        return current_smooth_time_adjusted, time_smooth_time_adjusted, current_multistep, time_multistep

    def offset_value(self, offset_test, *data):
        """
        Helper function shift the charging rates for the smooth current to ensure that both the multistep and
        smooth charging functions reach the end in the same length of time.

        Args:
            offset_test (float): Amount to shift all of the non-edge c rates in the smooth curve
            data (*args): ending time to hit, baseline charging rates to add to

        returns
        float: difference between the desired ending time and the actual ending time for the smooth curve
        """
        time_end_multistep, charging_c_rates, mesh_points = data
        soc_vector = np.linspace(self.soc_i, self.soc_f, self.soc_points)
        return time_end_multistep - self.shift_smooth_by_offset(offset_test,
                                                                charging_c_rates,
                                                                mesh_points)[0][soc_vector >= self.soc_f][0]

    def get_input_current_multistep_soc_as_x(self, charging_c_rates, mesh_points):
        """
        Helper function to generate a waveform with stepwise charging currents. Should return a vector
        of current values and vector of corresponding state of charge (soc) values.

        Args:
            charging_c_rates (list): c-rates for each of the charging steps. Each step is assumed to be an equal
                SOC portion of the charge, and the length of the list just needs to be at least 1
            mesh_points (list): soc values for beginning and end of each of the charging windows

        returns
        np.array: array with the current as a function of soc
        np.array: array with the corresponding soc values
        """
        soc_vector = np.linspace(self.soc_i, self.soc_f, self.soc_points)
        c_rate_i = np.zeros((len(mesh_points) - 1, len(soc_vector)))

        for i, elem in enumerate(mesh_points[0:-1]):
            soc1 = mesh_points[i]
            soc2 = mesh_points[i + 1]
            c_rate_i[i] = charging_c_rates[i] * np.heaviside(soc_vector - soc1, 0.5) * \
                np.heaviside(soc2 - soc_vector, 0.5)

            multistep_current_raw = sum(c_rate_i)

        c_rate_end = self.final_c_rate * np.heaviside(soc_vector - soc2, 0.5) * \
            np.heaviside(self.soc_f - soc_vector, 0.5)

        multistep_current = multistep_current_raw + c_rate_end

        return multistep_current, soc_vector

    def get_input_current_smooth_soc_as_x(self, charging_c_rates, mesh_points):
        """
        Helper function to generate a waveform with smoothly varying charging current. Should return a vector
        of current values and vector of corresponding state of charge (soc) values.

        Args:
            charging_c_rates (list): c-rates for each of the charging steps. Each step is assumed to be an equal
                SOC portion of the charge, and the length of the list just needs to be at least 1
            mesh_points (list): soc values for beginning and end of each of the charging windows

        returns
        np.array: array with the current as a function of soc
        np.array: array with the corresponding soc values
        """
        mesh_points_mid = np.copy(mesh_points)

        for indx in range(len(mesh_points) - 1):
            mesh_points_mid[indx] = (mesh_points[indx] + mesh_points[indx + 1]) / 2

        soc_vector = np.linspace(self.soc_i, self.soc_f, self.soc_points)

        mesh_points_mid = list([self.soc_i]) + list(mesh_points_mid) + list([self.soc_f * 1.01])
        mesh_points_mid = np.array(mesh_points_mid)

        charging_c_rate_start = charging_c_rates[0]
        rates = np.clip([charging_c_rate_start] + charging_c_rates + [charging_c_rates[-1]],
                        self.min_c_rate,
                        self.max_c_rate)

        interpolator = interpolate.PchipInterpolator(mesh_points_mid, rates, axis=0, extrapolate=0)

        charging_c_rate_soc1_end = interpolator.__call__(mesh_points[1])

        charging_c_rate_start = np.max(
            [charging_c_rates[0] - (charging_c_rate_soc1_end - charging_c_rates[0]), charging_c_rates[0]])
        charging_c_rate_start = np.min([self.max_c_rate, charging_c_rate_start])

        rates = np.clip([charging_c_rate_start] + charging_c_rates + [charging_c_rates[-1]],
                        self.min_c_rate,
                        self.max_c_rate)

        interpolator = interpolate.PchipInterpolator(mesh_points_mid, rates, axis=0, extrapolate=0)

        input_current = interpolator.__call__(soc_vector)
        input_current = np.nan_to_num(input_current, copy=False, nan=0)
        input_curent_smooth_soc_as_x = input_current

        return input_curent_smooth_soc_as_x, soc_vector

    def shift_smooth_by_offset(self, offset, charging_c_rates, mesh_points):
        """
        Helper function to shift the current values of the smooth charging curve to achieve the desired SOC sooner.

        Args:
            offset (float): Amount to shift all of the non-edge c rates in the smooth curve
            charging_c_rates (list): c-rates for each of the charging steps. Each step is assumed to be an equal
                SOC portion of the charge, and the length of the list just needs to be at least 1
            mesh_points (list): soc values for beginning and end of each of the charging windows

        returns
        np.array: array with the current as a function of soc
        np.array: array with the corresponding soc values
        """
        shifted_c_rates = [charging_c_rates[0]] + list(np.add(charging_c_rates[1:-1], offset)) + [charging_c_rates[-1]]
        # shifted_c_rates = list(np.add(charging_c_rates, offset))

        current_smooth_shifted, soc_vector = self.get_input_current_smooth_soc_as_x(shifted_c_rates, mesh_points)

        time_smooth_shifted = self.get_time_vector_from_c_vs_soc(soc_vector, current_smooth_shifted)

        return time_smooth_shifted, current_smooth_shifted

    @staticmethod
    def get_time_vector_from_c_vs_soc(soc_vector, current_soc_as_x):
        """
        Helper function to convert the soc vector to a time vector with the same datapoints

        Args:
            soc_vector (np.array): numpy array with
            current_soc_as_x (np.array): c-rates for all of the soc values

        returns
        np.array: array with the time corresponding to each of the current points
        """
        time_vector = np.cumsum((1 / current_soc_as_x) * np.gradient(soc_vector)) * 3600
        return time_vector



#!/usr/bin/env python3
#  Copyright (c) 2019 Toyota Research Institute

"""This script is a small utility to rename files and effectively move them from one project to another.

Usage:
    project_transfer.py [options]
    project_transfer.py (-h | --help)

Options:
    -h --help                Show this screen
    --version                Show version

"""

from botocore.exceptions import ClientError
import os
import shutil
import boto3
from monty.tempfile import ScratchDir


class ProjectTransfer:
    def __init__(self, input_project, output_project, bucket, prefix, dry_run=True):
        """
        Args:
            input_project (str): Name corresponding to the project to get the files from.
            output_project (str): Name of the project to transfer the files to.
            bucket (str): Name of the S3 bucket.
            prefix (str): Prefix of the input and output project in the S3 bucket.
            mode (str): mode to run in, if 'test' the output bucket is the 'beep-sync-test-stage' bucket.
        """

        self.input_project = input_project
        self.output_project = output_project
        self.bucket = bucket
        self.prefix = prefix
        self.dry_run = dry_run

    def get_list_files(self, excluded_string):
        all_objects = self.get_all_objects(self.bucket, self.prefix)
        object_names = [
            obj["Key"]
            for obj in all_objects
            if os.path.join(self.prefix, self.input_project) in obj["Key"]
        ]
        for string in excluded_string:
            object_names = [obj for obj in object_names if string not in obj]
        return object_names

    def get_all_objects(self, bucket, prefix):
        """
        Helper function to get common "subfolders" of folders in S3.

        Args:
            bucket (str): bucket name.
            prefix (str): prefix for which to list common prefixes.
        """
        object_names = []
        if not prefix.endswith("/"):
            prefix += "/"
        client = boto3.client("s3")
        paginator = client.get_paginator("list_objects_v2")
        result = paginator.paginate(Bucket=bucket, Delimiter="/", Prefix=prefix)
        for response in result:
            object_names = object_names + response["Contents"]
        return object_names

    def download_and_modify(self, object_names):
        """
        Downloads all S3 objects that are in the object_names list, runs a find and replace on the
        metadata in the first line of the file to change the project name from the input project name
        to the output project name.

        Args:
            object_names (list): List of object names in S3 that should be transferred to the output project.

        """
        with ScratchDir(".") as scratch_dir:
            for object_name in object_names:
                directory, name = os.path.split(object_name)
                s3 = boto3.client("s3")
                s3.download_file(
                    self.bucket, object_name, os.path.join(scratch_dir, name)
                )
                from_file = open(os.path.join(scratch_dir, name))
                line = from_file.readline()
                line = line.replace(self.input_project, self.output_project)
                to_file_name = os.path.join(
                    scratch_dir, name.replace(self.input_project, self.output_project)
                )
                to_file = open(to_file_name, "w")
                to_file.write(line)
                shutil.copyfileobj(from_file, to_file)
                if self.dry_run:
                    output_bucket = "beep-sync-test-stage"
                else:
                    output_bucket = self.bucket
                try:
                    new_object_name = os.path.join(
                        directory, name.replace(self.input_project, self.output_project)
                    )
                    s3.upload_file(to_file_name, output_bucket, new_object_name)
                except ClientError as e:
                    print(e)


if __name__ == "__main__":
    transfer = ProjectTransfer(
        "PredictionDiagnostics",
        "PreDiag",
        "beep-input-data-stage",
        "d3Batt/raw/maccor/STANFORD LOANER #2",
    )
    # List of strings here can be used to filer out specific file names so that not all files are transferred over
    # '_000052' is a file generated using an even earlier version of the protocol and probably should not be transferred
    # over
    excluded_strings = ["_000052", "Logs"]
    names = transfer.get_list_files(excluded_strings)
    transfer.download_and_modify(names)
    print(names)



import os
from setuptools import setup, find_packages
from pip._internal.req import parse_requirements
from pip._internal.network.session import PipSession

this_dir = os.path.dirname(os.path.abspath(__file__))
pip_requirements = parse_requirements(
    os.path.join(this_dir, "requirements.txt"), PipSession())
pip_requirements_test = parse_requirements(
    os.path.join(this_dir, "requirements-test.txt"), PipSession())

reqs = [pii.requirement for pii in pip_requirements]
reqs_test = [pii.requirement for pii in pip_requirements_test]

readme_path = os.path.join(this_dir, "README.md")

with open(readme_path, "r") as f:
    long_description = f.read()

# description must be one line
description = "beep is a python package supporting Battery Estimation and Early Prediction of battery cycle life."

setup(name="beep",
      url="https://github.com/TRI-AMDD/beep",
      version="2021.9.30.15",
      description=description,
      long_description=long_description,
      long_description_content_type='text/markdown',
      packages=find_packages(),
      install_requires=reqs,
      extras_require={"tests": reqs_test},
      entry_points={
          "console_scripts": [
              "collate = beep.collate:main",
              "validate = beep.validate:main",
              "structure = beep.structure.cli:main",
              "featurize = beep.featurize:main",
              "run_model = beep.run_model:main",
              "generate_protocol = beep.generate_protocol:main"
          ]
      },
      classifiers=[
          "Programming Language :: Python :: 3",
          "License :: OSI Approved :: Apache Software License",
          "Operating System :: OS Independent",
      ],
      package_data={
          "beep.conversion_schemas": ["*.yaml", "*.md"],
          "beep.protocol.biologic_templates": ["*.mps", "*.csv", "*.json"],
          "beep.protocol.procedure_templates": ["*.000", "*.csv", "*.json",
                                                "*.yaml"],
          "beep.protocol.protocol_schemas": ["*.yaml", "*.txt"],
          "beep.protocol.schedule_templates": ["*.sdu", "*.csv", "*.json"],
          "beep.validation_schemas": ["*.yaml"],
          "beep.model": ["*.model"],
          "beep.features": ["*.yaml"]

      },
      include_package_data=True,
      author="AMDD - Toyota Research Institute",
      author_email="patrick.herring@tri.global",
      maintainer="Patrick Herring",
      maintainer_email="patrick.herring@tri.global",
      license="Apache",
      keywords=[
          "materials", "battery", "chemistry", "science",
          "electrochemistry", "energy", "AI", "artificial intelligence"
      ],
      )



# Copyright [2020] [Toyota Research Institute]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Command line/batch interface for structuring many battery cycler runs.

Usage:
    structure [INPUT_JSON]

Options:
    -h --help       Show this screen
    --version       Show version


The `structure` script will run the data structuring on specified filenames corresponding
to validated raw cycler files.  It places the structured datafiles in `/data-share/structure`.

The input json must contain the following fields:
* `file_list` - a list of full path filenames which have been processed
* `validity` - a list of boolean validation results, e. g. `[True, True, False]`

The output json contains the following fields:

* `invalid_file_list` - a list of invalid files according to the validity
* `file_list` - a list of files which have been structured into processed_cycler_runs

Example:
```angular2
$ structure '{"validity": [false, false, true],
             file_list": ["/data-share/renamed_cycler_files/FastCharge/FastCharge_0_CH33.csv",
                          "/data-share/renamed_cycler_files/FastCharge/FastCharge_1_CH44.csv",
                          "/data-share/renamed_cycler_files/FastCharge/FastCharge_2_CH29.csv"]}''
{"invalid_file_list": ["/data-share/renamed_cycler_files/FastCharge/FastCharge_0_CH33.csv",
                       "/data-share/renamed_cycler_files/FastCharge/FastCharge_1_CH44.csv"],
 "file_list": ["/data-share/structure/FastCharge_2_CH29_structure.json"]}
```
"""

import re
import os
import json

from docopt import docopt
from monty.serialization import loadfn

from beep import logger, __version__
from beep.conversion_schemas import (
    FastCharge_CONFIG,
    xTesladiag_CONFIG,
    ARBIN_CONFIG,
    MACCOR_CONFIG,
    INDIGO_CONFIG,
    BIOLOGIC_CONFIG,
    NEWARE_CONFIG
)
from beep.structure.arbin import ArbinDatapath
from beep.structure.maccor import MaccorDatapath
from beep.structure.neware import NewareDatapath
from beep.structure.indigo import IndigoDatapath
from beep.structure.biologic import BiologicDatapath
from beep.structure.battery_archive import BatteryArchiveDatapath
from beep.structure.base import BEEPDatapath
from beep.utils import WorkflowOutputs
from beep.collate import add_suffix_to_filename


SERVICE_CONFIG = {"service": "DataStructurer"}


def process_file_list_from_json(file_list_json, processed_dir="data-share/structure/", omit_raw=True):
    """Function to take a json filename corresponding to a data structure
    with a 'file_list' and a 'validity' attribute, process each file
    with a corresponding True validity, dump the processed file into
    a predetermined directory, and return a jsonable dict of processed
    cycler run file locations

    Args:
        file_list_json (str): json string or json filename corresponding
            to a dictionary with a file_list and validity attribute,
            if this string ends with ".json", a json file is assumed
            and loaded, otherwise interpreted as a json string.
        processed_dir (str): location for processed cycler run output
            files to be placed.
        omit_raw (bool): Omit the raw_data from being saved to file. Creates
            legacy file structure for all structured datapaths.

    Returns:
        (str): json string of processed files (with key "processed_file_list").
            Note that this list contains None values for every file that
            had a corresponding False in the validity list.

    """
    # Get file list and validity from json, if ends with .json,
    # assume it's a file, if not assume it's a json string
    if file_list_json.endswith(".json"):
        file_list_data = loadfn(file_list_json)
    else:
        file_list_data = json.loads(file_list_json)

    # Setup workflow
    outputs = WorkflowOutputs()

    # Prepend optional root to output directory
    processed_dir = os.path.join(
        os.environ.get("BEEP_PROCESSING_DIR", "/"), processed_dir
    )

    if not os.path.exists(processed_dir):
        os.makedirs(processed_dir)

    file_list = file_list_data["file_list"]
    validities = file_list_data["validity"]
    run_ids = file_list_data["run_list"]
    processed_file_list = []
    processed_run_list = []
    processed_result_list = []
    processed_message_list = []
    invalid_file_list = []
    for filename, validity, run_id in zip(file_list, validities, run_ids):
        logger.info("run_id=%s structuring=%s", str(run_id), filename, extra=SERVICE_CONFIG)
        if validity == "valid":
            # Process datapath and dump to file

            dp = auto_load(filename)
            dp.autostructure()

            # raw_cycler_run = RawCyclerRun.from_file(filename)
            # processed_cycler_run = raw_cycler_run.to_processed_cycler_run()
            new_filename, ext = os.path.splitext(os.path.basename(filename))
            new_filename = new_filename + ".json"
            new_filename = add_suffix_to_filename(new_filename, "_structure")
            structured_run_loc = os.path.join(processed_dir, new_filename)
            structured_run_loc = os.path.abspath(structured_run_loc)
            dp.to_json_file(structured_run_loc, omit_raw)

            # Append file loc to list to be returned
            processed_file_list.append(structured_run_loc)
            processed_run_list.append(run_id)
            processed_result_list.append("success")
            processed_message_list.append({"comment": "", "error": ""})

        else:
            invalid_file_list.append(filename)

    output_json = {
        "file_list": processed_file_list,
        "run_list": processed_run_list,
        "result_list": processed_result_list,
        "message_list": processed_message_list,
        "invalid_file_list": invalid_file_list,
    }

    # Workflow outputs
    file_list_size = len(output_json["file_list"])
    if file_list_size > 1 or file_list_size == 0:
        logger.warning("{file_list_size} files being validated, should be 1")

    output_data = {
        "filename": output_json["file_list"][0],
        "run_id": output_json["run_list"][0],
        "result": output_json["result_list"][0],
    }

    outputs.put_workflow_outputs(output_data, "structuring")

    # Return jsonable file list
    return json.dumps(output_json)


def auto_load(filename):
    """Load any supported raw battery cycler file to the correct Datapath automatically.

    Matches raw file patterns to the correct datapath and returns the datapath object.

    Example:
        auto_load("2017-05-09_test-TC-contact_CH33.csv")

        >>> <ArbinDatapath object>

        auto_load("PreDiag_000287_000128short.092")

        >>> <MaccorDatapath object>

    Args:
        filename (str, Pathlike): string corresponding to battery cycler file filename.

    Returns:
        (beep.structure.base.BEEPDatapath): The datapath child class corresponding to this file.

    """
    if re.match(ARBIN_CONFIG["file_pattern"], filename) or re.match(FastCharge_CONFIG["file_pattern"], filename):
        return ArbinDatapath.from_file(filename)
    elif re.match(MACCOR_CONFIG["file_pattern"], filename) or re.match(xTesladiag_CONFIG["file_pattern"], filename):
        return MaccorDatapath.from_file(filename)
    elif re.match(INDIGO_CONFIG["file_pattern"], filename):
        return IndigoDatapath.from_file(filename)
    elif re.match(BIOLOGIC_CONFIG["file_pattern"], filename):
        return BiologicDatapath.from_file(filename)
    elif re.match(NEWARE_CONFIG["file_pattern"], filename):
        return NewareDatapath.from_file(filename)
    elif re.match(BatteryArchiveDatapath.FILE_PATTERN, filename):
        return BatteryArchiveDatapath.from_file(filename)
    else:
        raise ValueError("{} does not match any known file pattern".format(filename))


def auto_load_processed(path):
    """Load processed BEEP .json files regardless of their class.

    Enables loadfn capability for legacy BEEP files, since calling
    loadfn on legacy files will return dictionaries instead of objects
    or will outright fail.

    Examples:

        auto_load_processed("maccor_file_structured.json")

        >>> <MaccorDatapath object>

        auto_load_processed("neware_file_structured.json")

        >>> <NewareDatapath object>

        auto_load_processed("old_Biologic_ProcessedCyclerRun_legacy.json")

        >>> <BiologicDatapath object>

    Args:
        path (str, Pathlike): Path to the structured json file.

    Returns:
        (BEEPDatapath)

    """

    processed_obj = loadfn(path)

    if not isinstance(processed_obj, BEEPDatapath):
        if isinstance(processed_obj, dict):
            # default to Arbin file for BEEPDatapath purposes
            return ArbinDatapath.from_json_file(path)
        else:
            raise TypeError(f"Unknown type for legacy processed json! `{type(processed_obj)}`")
    else:
        return processed_obj


def main():
    """Main function of this module, takes in arguments of an input
    and output filename and uses the input file to create a
    structured data output for analysis/ML processing.
    """
    logger.info("starting", extra=SERVICE_CONFIG)
    logger.info("Running version=%s", __version__, extra=SERVICE_CONFIG)
    try:
        args = docopt(__doc__)
        input_json = args["INPUT_JSON"]
        print(process_file_list_from_json(input_json))
    except Exception as e:
        logger.error(str(e), extra=SERVICE_CONFIG)
        raise e
    logger.info("finish", extra=SERVICE_CONFIG)
    return None


if __name__ == "__main__":
    main()



# Copyright [2020] [Toyota Research Institute]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Module and scripts for training and predicting with models,
given a matching descriptor set.

Usage:
    run_model [INPUT_JSON] [--fit]

Options:
    -h --help        Show this screen
    --fit            <true_or_false>  [default: False] Fit model
    --version        Show version


The `run_model` script will generate a model and create predictions
based on the features previously generated by the featurize module.
It stores its outputs in `/data-share/predictions/`

The input json must contain the following fields
* `file_list` - list of files corresponding to model features

The output json will contain the following fields
* `file_list` - list of files corresponding to model predictions

Example:
$ run_model '{"file_list": ["/data-share/features/FastCharge_2_CH29_full_model_features.json"]}'
{
    "file_list": ["/data-share/predictions/FastCharge_2_CH29_full_model_predictions.json"]
}
"""
from __future__ import division
import os
import json
import pandas as pd
import numpy as np
import datetime
from docopt import docopt
from monty.json import MSONable
from monty.serialization import loadfn, dumpfn
from beep.collate import scrub_underscore_suffix, add_suffix_to_filename
from sklearn.linear_model import (
    Lasso,
    LassoCV,
    RidgeCV,
    Ridge,
    ElasticNetCV,
    ElasticNet,
    MultiTaskElasticNet,
    MultiTaskElasticNetCV,
)
from sklearn.metrics import mean_squared_error
from sklearn.model_selection import train_test_split
from beep import MODEL_DIR, logger, __version__

s = {"service": "DataAnalyzer"}
# Projects that have cycling profiles compatible with the FastCharge model should be included in the list below
DEFAULT_MODEL_PROJECTS = [
    "FastCharge",
    "ClosedLoopOED",
    "2017-05-12",
    "2017-06-30",
    "2018-04-12",
]
assert all("_" not in name for name in DEFAULT_MODEL_PROJECTS)


class DegradationModel(MSONable):
    """
    Object corresponding to prediction model.

    Attributes:
        name (str): instance name.
        model (dict): model specific parameters.
    """

    def __init__(self, name, model):
        """

        Args:
            name (str):
            model (dict):
        """
        self.name = name
        self.model = model

    @classmethod
    def from_name(cls, model_name="full_model"):
        """
        Args:
            model_name (str): name of method for featurization.
        """
        if model_name == "full_model":
            return cls.init_full_model()
        else:
            raise NotImplementedError

    @classmethod
    def from_serialized_model(
        cls, model_dir="data-share/model/", serialized_model=None
    ):
        """
        Class method to invoke from serialized model

        Args:
            model_dir (str): path to model directory
            serialized_model (str): name of model

        Returns:
            (DegradationModel): model object from stored model

        """
        if serialized_model is None:
            raise ValueError("Please specify model name stored in {}".format(model_dir))
        elif not os.path.exists(os.path.join(model_dir, serialized_model)):
            raise ValueError("Path invalid")
        else:
            trained_model = loadfn(os.path.join(model_dir, serialized_model))

        return cls(name=serialized_model.split(".")[0], model=trained_model)

    @classmethod
    def init_full_model(cls):
        """
        Predict using model coefficients generated by fitting to
        D3Batt early prediction manuscript data using BEEP codebase.

        Returns:
            DegradationModel
        """
        coefs = np.array(
            [
                77.64331966,
                -3.38357566,
                0.0,
                0.0,
                9.48943144,
                -0.0,
                -0.0,
                -0.0,
                -0.0,
                0.0,
                -172.1890952,
                -61.6947121,
                5.22729452,
                0.0,
                0.0,
                0.0,
                0.0,
                -31.84889315,
                -0.0,
                -0.0,
            ]
        )

        mu = np.array(
            [
                1.06823975e00,
                7.73847312e-01,
                1.07329200e00,
                1.13981637e06,
                6.34888946e02,
                -1.36155073e00,
                -1.72133649e00,
                -3.47084255e00,
                -1.22713574e00,
                2.63323211e-01,
                -1.47651181e00,
                3.73629665e01,
                2.95566653e01,
                -1.36504373e-05,
                1.07594232e00,
                -6.87024033e-05,
                1.08010237e00,
                1.67499372e-02,
                1.73228530e-02,
                -3.04463017e-04,
            ]
        )

        sigma = np.array(
            [
                1.41520865e-02,
                1.95217699e-01,
                1.59416725e-02,
                7.23228725e06,
                5.65537087e01,
                2.58001513e-01,
                2.70289386e-01,
                4.51079117e-01,
                6.71019154e-01,
                2.05683149e-01,
                2.25143929e-01,
                1.71008400e00,
                5.69541083e-01,
                5.98941337e-05,
                1.44721412e-02,
                6.14265444e-05,
                1.48826794e-02,
                6.54248192e-04,
                7.19974520e-04,
                5.86172005e-04,
            ]
        )

        model = {"coef_": coefs}
        name = "full_model"

        date_string = datetime.datetime.now()

        trained_model = {
            "model_type": "linear",
            "model": model,
            "confidence_bounds": 0.1,
            "regularization_type": "elasticnet",
            "timestamp": date_string.isoformat(),
            "dataset_id": None,
            "hyperparameters": {},
            "featureset_name": "full_model",
            "predicted_quantity": "cycle",
            "mu": np.array(mu),
            "sigma": np.array(sigma),
        }

        return cls(name, trained_model)

    def serialize(self, processed_dir="data-share/model"):
        """

        Args:
            processed_dir (dict): target directory.

        Returns:

        """
        if not os.path.exists(processed_dir):
            os.makedirs(processed_dir)
        dumpfn(self.model, os.path.join(processed_dir, self.name + ".model"))

    def as_dict(self):
        """
        Method for dictionary serialization.

        Returns:
            dict: corresponding to dictionary for serialization.

        """

        return {
            "@module": self.__class__.__module__,
            "@class": self.__class__.__name__,
            "name": self.name,
            "model": self.model,
        }

    @classmethod
    def from_dict(cls, d):
        """
        MSONable deserialization method.

        Args:
            d (dict):

        Returns:
            beep.run_model.DegradationModel:
        """
        return cls(**d)

    @classmethod
    def train(
        cls,
        list_of_featurized_jsons,
        dataset_id=None,
        model_type="linear",
        regularization_type="elasticnet",
        model_name="custom_model",
        hyperparameters=None,
    ):
        """
        Class method to accept a json string of featurized cycler run files
        and output model coefficients.  Trained models are serialized as
        a dictionary and pushed into a local file.

        Model coefficients are initialized after training.

        Args:
            list_of_featurized_jsons (str): json string of featurized cycler run files.
            dataset_id (str): unique_id corresponding to a list of run_ids that are used
                for model training.
            model_type (str): linear or random_forest.
            regularization_type (str): lasso or ridge or elasticnet
                (cv estimator chosen by default).
            model_name (str): custom name for the model.
            hyperparameters (dict): dictionary with the following attributes:
                random_state, test_size, k_fold, tol, l1_ratio.

        """
        if hyperparameters is None:
            hyperparameters = {
                "random_state": 2,
                "test_size": 0.3,
                "k_fold": 5,
                "tol": 0.0001,
                "l1_ratio": [0.1, 0.5, 0.7, 0.9, 0.95, 1],
                "max_iter": 1000000,
            }

        X, y, featureset_name, predicted_quantity = assemble_predictors(
            list_of_featurized_jsons
        )
        if model_type.lower() == "random_forest":
            (
                model,
                mu,
                s,
                relative_prediction_error,
                Rsquare,
                hyperparameters_optimized,
            ) = train_linear_model(X, y, **hyperparameters)
        elif model_type.lower() == "linear":
            (
                model,
                mu,
                s,
                relative_prediction_error,
                Rsquare,
                hyperparameters_optimized,
            ) = train_linear_model(X, y, **hyperparameters)
        else:
            raise NotImplementedError

        # Book-keeping
        date_string = datetime.datetime.now()
        trained_model = {
            "model_type": model_type.lower(),
            "model": model.__dict__,
            "confidence_bounds": relative_prediction_error,
            "Rsquare": Rsquare,
            "regularization_type": regularization_type,
            "timestamp": date_string.isoformat(),
            "dataset_id": dataset_id,
            "hyperparameters": hyperparameters_optimized,
            "featureset_name": featureset_name,
            "predicted_quantity": predicted_quantity,
        }
        if model_type.lower() == "linear":
            trained_model["mu"] = np.array(mu)
            trained_model["sigma"] = np.array(s)

        return cls(name=model_name, model=trained_model)

    def predict(self, features):
        """
        Args:
            features (beep.featurize.DegradationPredictor): features in
                DegradationPredictor format.

        Returns:
            prediction (float): float corresponding to predicted value.
        """
        if self.model["model_type"].lower() == "linear":
            X = (features.X - self.model["mu"]) / self.model["sigma"]
        else:
            X = features.X
        coefs = self.model["model"]["coef_"]
        prediction = (
            np.nansum((np.array(X) * coefs), axis=1) + self.model["model"]["intercept_"]
        )
        if prediction.ndim == 2:
            prediction = prediction.reshape((prediction.shape[1],))
        return prediction

    def prediction_to_dict(self, prediction, nominal_capacity=1.1):
        """
        Args:
            prediction (float or [float]): float or list of floats
                corresponding to predictions.
            nominal_capacity (float): Nominal capacity of the cell.

        Returns:
            dict: dictionary with predictions, error bars and model name.

        """
        output_dict = {}
        if len(prediction) == 1:
            output_dict["discharge_capacity"] = nominal_capacity * 0.8
            output_dict["fractional_error"] = self.model["confidence_bounds"]
        else:
            output_dict["discharge_capacity"] = (
                np.around(np.arange(0.98, 0.78, -0.03), 2) * nominal_capacity
            )

            # For now the API is only serving up a single value as the confidence interval
            # the model is producing MSE values for each of the prediction points, so the correct value over
            # the full set of prediction point should be the average of the MSE for each of the points
            # In the future the fractional errors can be an array, once the API and UI are ready for that
            # Taking the average is equivalent to
            output_dict["fractional_error"] = np.asarray(
                [np.average(self.model["confidence_bounds"])]
            )

        output_dict["cycle_number"] = prediction
        output_dict["model_type"] = self.model["model_type"]
        output_dict["predicted_quantity"] = self.model["predicted_quantity"]

        # Ideally, we will have a model_id, but using model_name for now.
        output_dict["model_name"] = self.name
        return output_dict


def train_linear_model(
    X,
    y,
    random_state=1,
    test_size=0.2,
    regularization_type="elasticnet",
    k_fold=5,
    max_iter=1000000,
    tol=0.0001,
    l1_ratio=None,
):
    """
    Function to train linear model with regularization and cross-validation.

    Args:
        X (pandas.DataFrame): dataframe of descriptors.
        y (pandas.DataFrame): dataframe of cycle lifetimes.
        random_state (int): seed for train/test split.
        test_size (float): proportion of the dataset reserved for model evaluation.
        regularization_type (str): lasso or ridge or elastic-net (with cv).
        k_fold (int): k in k-fold cross-validation.
        max_iter (int): maximum number of iterations for model fitting.
        tol (float): tolerance for optimization.
        l1_ratio ([float]): list of lasso to ridge ratios for elasticnet.

    Returns:
        sklearn.linear_model.LinearModel: fitted model.
        mu (float): Mean value of descriptors used in training.
        s (float): Std dev of descriptors used in training.

    """
    if l1_ratio is None:
        l1_ratio = [0.1, 0.5, 0.7, 0.9, 0.95, 1]
    X_train, X_test, y_train, y_test = train_test_split(
        X, y, test_size=test_size, random_state=random_state
    )

    # Standardize (training) data after train/test split
    mu = np.mean(X_train, axis=0)
    s = np.std(X_train, axis=0)
    X_scaled = (X_train - mu) / s
    hyperparameters = {
        "random_state": random_state,
        "test_size": test_size,
        "k_fold": k_fold,
        "tol": tol,
        "max_iter": max_iter,
    }
    if regularization_type == "lasso" and y.shape[1] == 1:
        lassocv = LassoCV(
            fit_intercept=True, alphas=None, tol=tol, cv=k_fold, max_iter=max_iter
        )
        lassocv.fit(X_scaled, y_train.values.ravel())
        # Set optimal alpha and refit model
        alpha_opt = lassocv.alpha_
        linear_model = Lasso(fit_intercept=True, alpha=alpha_opt, max_iter=max_iter)
        linear_model.fit(X_scaled, y_train.values)
        hyperparameters["l1_ratio"] = 1

    elif regularization_type == "ridge" and y.shape[1] == 1:
        ridgecv = RidgeCV(fit_intercept=True, alphas=None, cv=k_fold)
        ridgecv.fit(X_scaled, y_train.values.ravel())
        # Set optimal alpha and refit model
        alpha_opt = ridgecv.alpha_
        linear_model = Ridge(fit_intercept=True, alpha=alpha_opt)
        linear_model.fit(X_scaled, y_train)
        hyperparameters["l1_ratio"] = 0

    elif regularization_type == "elasticnet" and y.shape[1] == 1:
        elasticnetcv = ElasticNetCV(
            fit_intercept=True,
            normalize=False,
            alphas=None,
            cv=k_fold,
            l1_ratio=l1_ratio,
            max_iter=max_iter,
        )
        elasticnetcv.fit(X_scaled, y_train.values.ravel())

        # Set optimal alpha and l1_ratio. Refit model
        alpha_opt = elasticnetcv.alpha_
        l1_ratio_opt = elasticnetcv.l1_ratio_
        linear_model = ElasticNet(
            fit_intercept=True,
            normalize=False,
            l1_ratio=l1_ratio_opt,
            alpha=alpha_opt,
            max_iter=max_iter,
        )
        linear_model.fit(X_scaled, y_train)
        hyperparameters["l1_ratio"] = l1_ratio_opt

    # If more than 1 outcome present, perform multitask regression
    elif regularization_type == "elasticnet" and y.shape[1] > 1:
        multi_elasticnet_CV = MultiTaskElasticNetCV(
            fit_intercept=True,
            cv=k_fold,
            normalize=False,
            l1_ratio=l1_ratio,
            max_iter=max_iter,
        )
        multi_elasticnet_CV.fit(X_scaled, y_train)
        # Set optimal alpha and l1_ratio. Refit model
        alpha_opt = multi_elasticnet_CV.alpha_
        l1_ratio_opt = multi_elasticnet_CV.l1_ratio_
        linear_model = MultiTaskElasticNet(
            fit_intercept=True, normalize=False, max_iter=max_iter
        )
        linear_model.set_params(alpha=alpha_opt, l1_ratio=l1_ratio_opt)
        linear_model.fit(X_scaled, y_train)
        hyperparameters["l1_ratio"] = l1_ratio_opt
    else:
        raise NotImplementedError

    y_pred = linear_model.predict((X_test - mu) / s)
    Rsq = linear_model.score((X_test - mu) / s, y_test)
    # Compute 95% confidence interval
    # Multioutput = 'raw_values' provides prediction error per output
    pred_actual_ratio = [x / y for x, y in zip(y_pred, np.array(y_test))]
    relative_prediction_error = 1.96 * np.sqrt(
        mean_squared_error(
            np.ones(y_pred.shape), pred_actual_ratio, multioutput="raw_values"
        )
        / y_pred.shape[0]
    )
    hyperparameters["alpha"] = alpha_opt
    return linear_model, mu, s, relative_prediction_error, Rsq, hyperparameters


def assemble_predictors(file_list_json):
    """
    Method to assemble predictor dataframe from a json string of paths to feature vectors.

    Args:
        file_list_json (str): json string corresponding to a dictionary
            with a file_list attribute.

    Returns:
        pandas.DataFrame: Dataframe of size (n,m). n = number of cells, m = number of features.
        pandas.Series: Series of length n.
    """
    if file_list_json.endswith(".json"):
        file_list_data = loadfn(file_list_json)
    else:
        file_list_data = json.loads(file_list_json)

    X = pd.DataFrame()
    y = pd.DataFrame()

    for path in file_list_data["file_list"]:
        features = loadfn(path)
        X = X.append(features.X)
        if isinstance(features.y, (int, float)):
            y = y.append(pd.DataFrame([features.y]))
        else:
            y = y.append(pd.DataFrame(features.y))

    # Most NaNs should be handled at featurization, but if any crop-up during
    # model fitting, impute missing values with median of that feature.
    X = X.apply(lambda x: x.fillna(x.median()), axis=0)
    y = y.apply(lambda x: x.fillna(x.median()), axis=0)

    return X, y, features.name, features.predicted_quantity


def add_file_prefix_to_path(path, prefix):
    """
    Helper function to add file prefix to path

    Args:
        path (str): full path to file
        prefix (str): prefix for file

    Returns:
        str: path with prefix appended to filename

    """
    split_path = list(os.path.split(path))
    split_path[-1] = prefix + split_path[-1]
    return os.path.join(*split_path)


def get_project_name_from_list(file_list):
    """
    Helper function to get the project name from a list of file names

    Args:
        file_list (list): List of full file names (paths).
    Returns:
        str: Name of the project.
    """
    names_list = []
    for file in file_list:
        _, file_tail = os.path.split(file)
        names_list.append(file_tail.split("_")[0])
    project_name = max(set(names_list), key=names_list.count)

    return project_name


def process_file_list_from_json(
    file_list_json,
    model_dir="/data-share/models/",
    processed_dir="data-share/predictions/",
    hyperparameters=None,
    model_name=None,
    predict_only=True,
):
    """
    Function to take a json file containing featurized json locations,
    train a new model if necessary, write files containing predictions into a
    predetermined directory, and return a jsonable dict of prediction file locations

    Args:
        file_list_json (str): json string or json filename corresponding
            to a dictionary with a file_list attribute,
            if this string ends with ".json", a json file is assumed
            and loaded, otherwise interpreted as a json string
        model_dir (str): location where models are serialized and stored
        processed_dir (str): location for processed cycler run output files
            to be placed
        hyperparameters (dict): dictionary of hyperparameters to optimize/use for training
        model_name (str): name of feature generation method
        predict_only (bool):

    Returns:
        str: json string of feature files (with key "feature_file_list").

    """
    # Get file list and validity from json, if ends with .json,
    # assume it's a file, if not assume it's a json string
    if file_list_json.endswith(".json"):
        file_list_data = loadfn(file_list_json)
    else:
        file_list_data = json.loads(file_list_json)

    # Setup workflow TODO

    # Add BEEP_PROCESSING_DIR to processed_dir
    processed_dir = os.path.join(
        os.environ.get("BEEP_PROCESSING_DIR", "/"), processed_dir
    )
    if not os.path.exists(processed_dir):
        os.makedirs(processed_dir)

    file_list = file_list_data["file_list"]
    run_ids = file_list_data["run_list"]
    processed_run_list = []
    processed_result_list = []
    processed_message_list = []
    processed_paths_list = []
    project_name = get_project_name_from_list(file_list)
    if predict_only:
        features = loadfn(file_list[0])
        if model_name is None and project_name in DEFAULT_MODEL_PROJECTS:

            if features.prediction_type == "multi":
                model = DegradationModel.from_serialized_model(
                    model_dir=model_dir, serialized_model="d3batt_multi_point.model"
                )
            else:
                model = DegradationModel.from_serialized_model(
                    model_dir=model_dir, serialized_model="d3batt_single_point.model"
                )

        elif model_name is None and project_name not in DEFAULT_MODEL_PROJECTS:
            output_data = {
                "file_list": [],
                "run_list": [],
                "result_list": [],
                "message_list": [],
            }

            # Return jsonable file list
            return json.dumps(output_data)

        else:
            model = DegradationModel.from_serialized_model(
                model_dir=model_dir, serialized_model=model_name
            )

    else:
        if hyperparameters is None:
            hyperparameters = {
                "random_state": 1,
                "test_size": 0.3,
                "k_fold": 5,
                "tol": 0.001,
                "l1_ratio": [0.1, 0.5, 0.7, 0.9, 0.95, 0.99, 1],
            }

        dataset_id = file_list_data.get("dataset_id")
        model = DegradationModel.train(
            file_list_json,
            dataset_id=dataset_id,
            model_type="linear",
            regularization_type="elasticnet",
            model_name=model_name,
            hyperparameters=hyperparameters,
        )
        logger.warning("fitting=%s dataset=%s", model.name, str(dataset_id), extra=s)

    for path, run_id in zip(file_list, run_ids):
        logger.info(
            "model=%s run_id=%s predicting=%s", model.name, str(run_id), path, extra=s
        )
        features = loadfn(path)
        prediction = model.predict(features)
        prediction_dict = model.prediction_to_dict(
            prediction, features.nominal_capacity
        )
        new_filename = os.path.basename(path)
        new_filename = scrub_underscore_suffix(new_filename)
        new_filename = add_suffix_to_filename(new_filename, "_predictions")
        processed_path = os.path.join(processed_dir, new_filename)
        processed_path = os.path.abspath(processed_path)
        dumpfn(prediction_dict, processed_path)

        # Append file loc to list to be returned
        processed_paths_list.append(processed_path)
        processed_run_list.append(run_id)
        processed_result_list.append("success")
        processed_message_list.append({"comment": "", "error": ""})

    output_data = {
        "file_list": processed_paths_list,
        "run_list": processed_run_list,
        "result_list": processed_result_list,
        "message_list": processed_message_list,
    }

    # Return jsonable file list
    return json.dumps(output_data)


def main():
    """
    Main function of this module, takes in arguments of an input
    and output filename corresponding to featurized run data
    and creates a predictor object output for analysis/ML processing
    """
    # Parse args and construct initial cycler run
    logger.info("starting", extra=s)
    logger.info("Running version=%s", __version__, extra=s)
    try:
        args = docopt(__doc__)
        input_json = args["INPUT_JSON"]
        if args["--fit"]:
            print(
                process_file_list_from_json(
                    input_json, predict_only=False, model_dir=MODEL_DIR
                ),
                end="",
            )
        else:
            print(process_file_list_from_json(input_json, model_dir=MODEL_DIR), end="")
    except Exception as e:
        logger.error(str(e), extra=s)
        raise e
    logger.info("finish", extra=s)
    return None


if __name__ == "__main__":
    print(main())
